# -*- coding: utf-8 -*-

# backend.py
# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import gtk, gobject, os, subprocess, time, re
import const

# D-BUS specific xdvi parameters
DBUS_INCANTATION = "\"dbus-send --type=method_call --dest=org.gedit.LaTeXPlugin /org/gedit/DBUSInterface org.gedit.LaTeXPluginIFace.inverse string:%f int32:%l\""

SUFFIXES = {const.PDF : ".pdf", const.DVI : ".dvi", const.PS : ".ps", const.PNG : ".png", const.JPG : ".jpg", const.GIF : ".gif"}

class RubberBackend(gobject.GObject):
	""" This encapsulates the LaTeX tool chain. A Dialog may be connected to the
		signals to visualize the progress """
	__gsignals__ = {
		# def callback(backend, totalSteps)
		"invoked" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ([gobject.TYPE_INT])),
		# def callback(backend, fractionDone, currentJob)
		"progress" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ([gobject.TYPE_FLOAT, gobject.TYPE_STRING])),
		"finished" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
		# def callback(backend, fileName, lineNumber, errorMessage)
		"failed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ([gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_STRING]))
	}
	
	_filename, _filenameTrunc = None, None		# input filename
	_format = None
	_batch = []					# batch containing (command, message)-tuples
	_batchPointer = 0			# current batch entry beeing processed
	
	_process = None
	_exitWatch = None
	_timestamp = None			# timestamp when the compiler has been started
	
	_imagePNGMode = "alpha"		# PNG mode (-sDEVICE=png%s)
	_imageAA = 4				# antialiasing factor
	_imageRes = 120				# resolution in dpi
	_imageBox = False			# use a bounding box instead of the entire page
	
	_cleanupQueue = []
	details = []				# detailed (line, error)-tuples
	cacheEnabled = True
	
	def __init__(self):
		gobject.GObject.__init__(self)
	
	def compile(self, format, filename=None, string=None):
		## FIXME: raise an exception if we fail here so that __init__ can
		## determine whether to enable DVI search etc.
		self._format = format
		if filename:
			if open(filename).read().find(r"\begin{document}") == -1:
				# the document is not a adequate document and therefore included by a master
				# find that parent document...
				masterFilename = self._findMaster(filename)
				if masterFilename:
					self._filename = masterFilename
				else:
					return False		# no master found
			else:
				self._filename = filename
		elif string:
			import tempfile
			tempObj = tempfile.mkstemp(".tex")
			tempFile = os.fdopen(tempObj[0], "w")
			tempFile.write(string)
			tempFile.close()
			tempFilename = tempObj[1]
			self._cleanupQueue.append(tempFilename)
			self._filename = tempFilename
		self._filenameTrunc = os.path.splitext(self._filename)[0]
		self._assembleBatch()
		self._batchPointer = 0
		self._timestamp = int(time.time())
		self._aborted = False
		self.emit("invoked", len(self._batch))
		os.chdir(os.path.dirname(self._filename))
		self._runBatch()
		
		return True
	
	def _findMaster(self, childFilename):
		""" Find the document that includes the one to be compiled """
		import glob
		docName = os.path.splitext(os.path.basename(childFilename))[0]
		for filename in glob.glob(os.path.dirname(childFilename) + "/*.tex"):
			contents = open(filename).read()
			if contents.find(r"\include{" + docName) != -1:		# this one includes the file
				if contents.find(r"\begin{document}") != -1:	# and it's a full document >> bingo!
					return filename
				else:					# it's not a full document either, not found
					return None
		return None					# not found in this directory
	
	def findLine(self, lineNumber):
		""" Instructs xdvi to mark the area corresponding to a line in the source """
		os.system("xdvi -editor %s -sourceposition \"%s%s\" &" % (DBUS_INCANTATION, lineNumber, self._filename))
	
	def abort(self):
		self._aborted = True
		if self._process:
			import signal
			gobject.source_remove(self._exitWatch)
			os.kill(self._process.pid, signal.SIGTERM)
			self._process, self._exitWatch = None, None
			self.emit("finished")
	
	def setImageParameters(self, pngMode="alpha", aaFactor=4, resolution=120, boundingBox=False):
		self._imagePNGMode, self._imageAAFactor, self._imageResolution, self._imageBox = pngMode, aaFactor, resolution, boundingBox
	
	def _assembleBatch(self):
		# invoke rubber with the special module enabling source specials
		rubberStub = "rubber --force --short --quiet --inplace"
		
		if self._format == const.PDF:
			self._batch = [("%s --pdf \"%s\"" % (rubberStub, self._filename), "Compiling source to PDF")]
		elif self._format == const.DVI:
			self._batch = [("%s -m %s/latex-src \"%s\"" % (rubberStub, const.PATH, self._filename), "Compiling source to DVI")]
		elif self._format == const.PS:
			self._batch = [("%s --ps \"%s\"" % (rubberStub, self._filename), "Compiling source to PostScript")]
		elif self._format in [const.PNG, const.JPG, const.GIF]:
			rubber = ("%s \"%s\"" % (rubberStub, self._filename), "Compiling source to DVI")
			if self._imageBox:
				if self._format == const.PNG:
					mode = "png" + self._imagePNGMode
				elif self._format == const.JPG:
					mode = "jpeg"
				elif self._format == const.GIF:
					mode = "gif"
				self._batch = [
					rubber,
					("dvips -D %s -q -E -o \"%s.ps\" \"%s.dvi\"" % (self._imageRes, self._filenameTrunc, self._filenameTrunc), "Converting DVI to PostScript"),
					("%s/eps2png -%s -resolution=%s -aa=%s \"%s.eps\"" % (const.PATH, mode, self._imageRes, self._imageAA, self._filenameTrunc), "Rendering PostScript to PNG")
				]
			else:
				dvips = ("dvips -D %s -q -o \"%s.ps\" \"%s.dvi\"" % (self._imageRes, self._filenameTrunc, self._filenameTrunc), "Converting DVI to PostScript")
				if self._format == const.PNG:
					self._batch = [
						rubber,
						dvips,
						("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=png%s -sOutputFile=%s.png %s.ps quit.ps"
							% (self._imageRes, self._imageAA, self._imageAA, self._imagePNGMode, self._filenameTrunc, self._filenameTrunc), "Rendering PostScript to PNG")
					]
				elif self._format == const.JPG:
					self._batch = [
						rubber,
						dvips,
						("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=jpeg -sOutputFile=%s.jpg %s.ps quit.ps"
							% (self._imageRes, self._imageAA, self._imageAA, self._filenameTrunc, self._filenameTrunc), "Rendering PostScript to JPEG")
					]
				elif self._format == const.GIF:
					self._batch = [
						rubber,
						dvips,
						("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=ppm -sOutputFile=%s.ppm %s.ps quit.ps"
							% (self._imageRes, self._imageAA, self._imageAA, self._filenameTrunc, self._filenameTrunc), "Rendering PostScript to PPM"),
						("ppmtogif %s.ppm > %s.gif" % (self._filenameTrunc, self._filenameTrunc), "Converting PPM to GIF")
					]
			
	def _runBatch(self):
		command = self._batch[self._batchPointer][0]
		message = self._batch[self._batchPointer][1]
		fraction = float(self._batchPointer) / float(len(self._batch))
		self.emit("progress", fraction, message)
		self._process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		self._exitWatch = gobject.child_watch_add(self._process.pid, self._onExit)
	
	def _onExit(self, pid, condition):
		gobject.source_remove(self._exitWatch)
		self._process, self._exitWatch = None, None
		if condition:
			# read detailed error messages
			output = os.popen("rubber-info --errors %s 2>&1 && rubber-info --refs %s 2>&1" % (self._filename, self._filename)).read()
			self.details = []
			for line in output.splitlines():
				parts = line.split(":")
				if len(parts) == 1 and len(line):
					filename, lineNumber, message = "", -1, line.lstrip()
				elif len(parts) == 2:
					filename, lineNumber, message = os.path.basename(parts[0]), -1, parts[1].lstrip()
				elif len(parts) >= 3:
					filename, lineNumber, message = os.path.basename(parts[0]), parts[1], ("".join(parts[2:])).lstrip()
				else:
					continue
				self.details.append((filename, int(lineNumber), message))
			
			# return the first error
			if len(self.details):
				self.emit("failed", self.details[0][0], self.details[0][1], self.details[0][2])
			else:
				self.emit("failed", "", -1, "")
			
			self._cleanup()
			self._viewOutput()
		else:
			if self._batchPointer < len(self._batch) - 1:
				self._batchPointer += 1
				self._runBatch()
			else:
				self.emit("finished")
				self._cleanup()
				self._viewOutput()
	
	def _viewOutput(self):
		""" Open the output file in a viewer """
		suffix = SUFFIXES[self._format]
		try:
			mtime = os.path.getmtime("%s%s" % (self._filenameTrunc, suffix))
		except OSError:
			# file not found
			return
		if mtime < self._timestamp:
			# file too old
			return
		if self._format == const.PDF:
			os.system("evince \"%s.pdf\" &" % self._filenameTrunc)
		elif self._format == const.DVI:
			os.system("xdvi -unique -s 6 -bg white -editor %s \"%s.dvi\" &" % (DBUS_INCANTATION, self._filenameTrunc))
		elif self._format == const.PS:
			os.system("evince \"%s.ps\" &" % self._filenameTrunc)
		elif self._format in [const.PNG, const.JPG, const.GIF]:
			#os.system("eog \"%s%s\" &" % (self._filenameTrunc, suffix))
			os.system("gnome-open \"%s%s\" &" % (self._filenameTrunc, suffix))
	
	def _cleanup(self):
		""" Delete files enqueued for cleanup """
		for filename in self._cleanupQueue:
			try:
				os.unlink(filename)
			except OSError:
				pass

gobject.type_register(RubberBackend)


IDLE = 0
RUNNING = 1
FAILED = 2

class BackendDialog(gtk.Dialog):
	_state = IDLE
	
	def __init__(self, backend, instance):
		self._instance = instance
		self._backend = backend
		# create
		gtk.Dialog.__init__(self, "Compile", None,
				gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT, (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
		self._labelHeader = gtk.Label()
		self._labelError = gtk.Label()
		self._progress = gtk.ProgressBar()
		self._expander = gtk.Expander("<b>More _Detailed Errors</b>")
		self._expander.set_use_markup(True)
		self._expander.set_use_underline(True)
		self._storeDetails = gtk.ListStore(str, str, str)	# filename, str(line), error
		viewDetails = gtk.TreeView(self._storeDetails)
		viewDetails.insert_column_with_attributes(-1, _("Filename"), gtk.CellRendererText(), text=0)
		viewDetails.insert_column_with_attributes(-1, _("Line"), gtk.CellRendererText(), text=1)
		viewDetails.insert_column_with_attributes(-1, _("Message"), gtk.CellRendererText(), text=2)
		viewDetails.set_headers_visible(False)
		viewDetails.show()
		viewDetails.connect("row-activated", self.onErrorSelected)
		scrollDetails = gtk.ScrolledWindow()
		scrollDetails.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
		scrollDetails.set_shadow_type(gtk.SHADOW_IN)
		scrollDetails.add(viewDetails)
		scrollDetails.set_size_request(-1, 100)
		scrollDetails.show()
		self._expander.add(scrollDetails)
		# pack
		box = gtk.VBox()
		box.set_border_width(5)
		box.set_spacing(5)
		box.add(self._labelHeader)
		box.add(self._progress)
		box.add(self._labelError)
		box.add(self._expander)
		box.show()
		self.vbox.add(box)
		# configure
		self.set_resizable(False)
		self.set_border_width(5)
		self.set_has_separator(False)
		self.vbox.set_spacing(10)
		self._labelHeader.set_alignment(0, .5)
		self._labelHeader.set_size_request(300, -1)
		self._labelHeader.show()
		self._labelError.set_alignment(0, .5)
		self._progress.show()
		# connect
		backend.connect("invoked", self._onBackendInvoked)
		backend.connect("progress", self._onBackendProgress)
		backend.connect("failed", self._onBackendFailed)
		backend.connect("finished", self._onBackendFinished)
		self.connect("response", self._onResponse)
		self.connect("delete_event", self._onDeleteEvent)
	
	def onErrorSelected(self, treeview, path, column):
		liststore = treeview.get_model()
		line = liststore.get_value(liststore.get_iter(path), 1)
		# jump to line
		if line.isdigit():
			self._instance.get_document().scrollToLine(int(line))
	
	def _onDeleteEvent(self, widget, event):
		self._switchState(IDLE)
		return True
	
	def _switchState(self, state, showProgress=False):
		self._state = state
		if state == IDLE:
			self.hide()
			self._storeDetails.clear()
		elif state == RUNNING:
			self.action_area.get_children()[0].set_label(gtk.STOCK_CANCEL)
			self._labelError.hide()
			self._expander.hide()
			self._progress.set_fraction(0)
			if showProgress:
				self._progress.show()
			else:
				self._progress.hide()
			self.show()
		elif state == FAILED:
			self.action_area.get_children()[0].set_label(gtk.STOCK_CLOSE)
			self._labelError.show()
			self._expander.show()
	
	def _onResponse(self, dialog, response):
		if self._state == RUNNING:
			self._backend.abort()
		elif self._state == FAILED:
			self._switchState(IDLE)
	
	def _onBackendInvoked(self, backend, totalSteps):
		self._switchState(RUNNING, bool(totalSteps > 1))
	
	def _onBackendProgress(self, backend, fraction, job):
		self._labelHeader.set_markup("<big><b>%s</b></big>" % job)
		self._progress.set_fraction(fraction)
	
	def _onBackendFinished(self, backend):
		self._switchState(IDLE)
	
	def _onBackendFailed(self, backend, filename, line, error):
		if len(error):
			if line > -1:
				self._labelError.set_markup("Error in line %s: <i>%s</i>" % (line, error))
			else:
				self._labelError.set_markup("<i>%s</i>" % error)
		else:
			self._labelError.set_markup("<i>Unknown error</i>")
		self._storeDetails.clear()
		for item in self._backend.details:
			if item[1] > -1:
				lineNumber = str(item[1])
			else:
				lineNumber = ""
			self._storeDetails.append([item[0], lineNumber, item[2]])
		self._switchState(FAILED)
		# jump to line
		if line > -1:
			self._instance.get_document().scrollToLine(line)
		
