# -*- coding: utf-8 -*-

# config.py
# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import gtk, gconf

_GCONF_APP_PATH = "/apps/gedit-latex-plugin"

class Preferences:
	"""
	Encapsulates a GConf connection
	"""
	def __init__(self):
		self._client = gconf.client_get_default()
		self._client.add_dir(_GCONF_APP_PATH, gconf.CLIENT_PRELOAD_NONE)
		
		# work-around for the changed type of these keys:
		# unset them to avoid type mismatches
		if type(self.__getitem__("/last-document-class")) is int:
			self.unset("/last-document-class")
		if type(self.__getitem__("/last-paper-size")) is int:
			self.unset("/last-paper-size")
		
		# add the Schrödinger equation as an example for user-definded macros
		schroedinger_dir = "/user-defined-macros/cf7d1a74a05b5c6d56e0fa245cf0f33f"
		if not self.dir_exists(schroedinger_dir):
			self.add_dir(schroedinger_dir)
			self.__setitem__(schroedinger_dir + "/name", "Schrödinger")
			self.__setitem__(schroedinger_dir + "/scope", 2)
			self.__setitem__(schroedinger_dir + "/definition", "\\mathrm{i}\\cdot\\hbar\\cdot\\frac{\\partial}{\\partial t}\\psi(\\mathbf{r},t) \\;=\\; - \\frac{\\hbar^2}{2m}\\cdot\\nabla^2\\psi(\\mathbf{r},t) + V(\\mathbf{r},t)\\cdot\\psi(\\mathbf{r},t)")
			
	def __getitem__(self, key):
		"""
		Returns the value of a key
		"""
		value = self._client.get(_GCONF_APP_PATH + key)
		if not value:
			return None
		
		type = value.type.value_name
		if type == "GCONF_VALUE_STRING":
			return value.get_string()
		elif type == "GCONF_VALUE_INT":
			return value.get_int()
		elif type == "GCONF_VALUE_BOOL":
			return value.get_bool()
		elif type == "GCONF_VALUE_LIST":
			return [v.get_string() for v in value.get_list()]
				
	def __setitem__(self, key, value):
		vtype = type(value)
		if vtype == str:
			self._client.set_string(_GCONF_APP_PATH + key, value)
		elif vtype == int:
			self._client.set_int(_GCONF_APP_PATH + key, value)
		elif vtype == bool:
			self._client.set_bool(_GCONF_APP_PATH + key, value)
		elif vtype == list:
			self._client.set_list(_GCONF_APP_PATH + key, gconf.VALUE_STRING, [str(v) for v in value])
	
	def setdefault(self, key, default):
		"""
		Similar to __getitem__(), but returns <default> and assigns it to <key>
		if it doesn't exist
		"""
		if self._client.get_without_default(_GCONF_APP_PATH + key):
			return self.__getitem__(key)
		else:
			self.__setitem__(key, default)
			return default
	
	def unset(self, key):
		"""
		Unsets (removes) an entry 
		(if all entries in a directory are unset, then this directory is removed)
		"""
		self._client.unset(_GCONF_APP_PATH + key)
	
	def get_subdirs(self, key):
		"""
		Returns all directories below <key>
		"""
		return self._client.all_dirs(_GCONF_APP_PATH + key)
	
	def add_dir(self, key):
		"""
		Adds a directory
		"""
		self._client.add_dir(_GCONF_APP_PATH + key, gconf.CLIENT_PRELOAD_NONE)
	
	def dir_exists(self, key):
		"""
		Returns True if the directory exists, False otherwise
		"""
		return self._client.dir_exists(_GCONF_APP_PATH + key)
	
	def remove_dir(self, key):
		"""
		"""
		print "removing " + _GCONF_APP_PATH + key
		self._client.remove_dir(_GCONF_APP_PATH + key)
	
	def add_notify(self, callback, subdir=""):
		"""
		Adds a notify event handler for a specific subdirectory or the main app
		path
		"""
		self._client.notify_add(_GCONF_APP_PATH + subdir, callback)

class ConfigDialog(gtk.Dialog):
	def __init__(self):
		gtk.Dialog.__init__(self, _("Configure LaTeX Plugin"), None, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT, 
				(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))
		self._prefs = Preferences()
		self.set_resizable(False)
		self.set_border_width(5)
		self.set_has_separator(False)
		self.vbox.set_spacing(10)
		
		# rubber Cache
		
		#self.checkCache = gtk.CheckButton(_("Enable the experimental cache system to speed up subsequent compilations"))
		#self.checkCache.set_active(self._prefs.setdefault("/rubber-cache-enabled", True))
		
		#algCache = gtk.Alignment(.5, .5, 1, 1)
		#algCache.set_padding(5, 0, 15, 0)
		#algCache.add(self.checkCache)
		
		#frmCache = gtk.Frame(_("<b>Compiling Backend</b>"))
		#frmCache.get_label_widget().set_use_markup(True)
		#frmCache.set_shadow_type(gtk.SHADOW_NONE)
		#frmCache.add(algCache)
		
		#self.vbox.add(frmCache)
		
		# Auto-Completion / Auto-Closing
		
		self.checkCompletion = gtk.CheckButton(_("Enable automatic completion"))
		self.checkCompletion.set_active(self._prefs.setdefault("/auto-completion-enabled", True))
		self.checkClosing = gtk.CheckButton(_("Enable automatic closing of environment commands"))
		self.checkClosing.set_active(self._prefs.setdefault("/auto-closing-enabled", True))
		
		vboxInput = gtk.VBox()
		vboxInput.add(self.checkCompletion)
		vboxInput.add(self.checkClosing)
		
		algInput = gtk.Alignment(.5, .5, 1, 1)
		algInput.set_padding(5, 0, 15, 0)
		algInput.add(vboxInput)
		
		frmInput = gtk.Frame(_("<b>Input Assistance</b>"))
		frmInput.get_label_widget().set_use_markup(True)
		frmInput.set_shadow_type(gtk.SHADOW_NONE)
		frmInput.add(algInput)
		
		self.vbox.add(frmInput)
		
		# DVI source specials
		
		self.checkDVI = gtk.CheckButton(_("Enable forward/inverse search in DVI documents via D-BUS"))
		self.checkDVI.set_active(self._prefs.setdefault("/dvi-search-enabled", False))
		
		algDVI = gtk.Alignment(.5, .5, 1, 1)
		algDVI.set_padding(5, 0, 15, 0)
		algDVI.add(self.checkDVI)
		
		frmDVI = gtk.Frame(_("<b>DVI Source Specials</b>"))
		frmDVI.get_label_widget().set_use_markup(True)
		frmDVI.set_shadow_type(gtk.SHADOW_NONE)
		frmDVI.add(algDVI)
		
		self.vbox.add(frmDVI)
		
		# DPI
		
		lblResolution = gtk.Label(_("Screen Resolution:"))
		lblResolution.set_alignment(0, .5)
		self.spinDPI = gtk.SpinButton(gtk.Adjustment(self._prefs.setdefault("/dpi", 96), 1, 1000, 1), 1)
		lblResolutionU = gtk.Label("dpi")
		lblResolutionU.set_alignment(0, .5)
		
		hboxResolution = gtk.HBox()
		hboxResolution.set_spacing(5)
		hboxResolution.pack_start(lblResolution, False, False)
		hboxResolution.pack_start(self.spinDPI, False, False)
		hboxResolution.pack_start(lblResolutionU, False, False)
		
		algResolution = gtk.Alignment(.5, .5, 1, 1)
		algResolution.set_padding(5, 0, 15, 0)
		algResolution.add(hboxResolution)
		
		frmResolution = gtk.Frame(_("<b>Unit Conversion</b>"))
		frmResolution.get_label_widget().set_use_markup(True)
		frmResolution.set_shadow_type(gtk.SHADOW_NONE)
		frmResolution.add(algResolution)
		
		self.vbox.pack_start(frmResolution, False, False)
		
		self.connect("response", self.onResponse)
		self.set_default_response(gtk.RESPONSE_OK)
		
		self.show_all()
	
	def onResponse(self, dialog, response):
		if response == gtk.RESPONSE_OK:
			#self._prefs["/rubber-cache-enabled"] = self.checkCache.get_active()
			self._prefs["/auto-completion-enabled"] = self.checkCompletion.get_active()
			self._prefs["/auto-closing-enabled"] = self.checkClosing.get_active()
			self._prefs["/dvi-search-enabled"] = self.checkDVI.get_active()
			self._prefs["/dpi"] = self.spinDPI.get_value_as_int()
		self.destroy()
